# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
        name            = 'pygrametlsilex',
        version         = '0.0.1',
        author          = 'rvnovaes',
        author_email    = 'roberto.novaes@silexsistemas.com.br',
        url             = 'https://bitbucket.org/silex/pygrametlsilex',
        description     = 'ETL (Extract, Transform, Load) written in Python.',
        license         = 'BSD',
        packages        = ['pygrametlsilex'],
        classifiers     = ['Development Status :: 3 - Alpha',
                           'Environment :: Console',
                           'Intended Audience :: Developers',
                           'License :: OSI Approved :: BSD License',
                           'Operating System :: OS Independent',
                           'Programming Language :: Python',
                           'Programming Language :: Python :: 2',
                           'Programming Language :: Python :: 2.7',
                           'Topic :: Software Development :: Libraries :: Python Modules',
                           'Topic :: Database',
                           'Topic :: Utilities'],
        )